/*  Copyright (C) 1995  Byron Stanoszek
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 **************************************************************************/

/* snddriv.c */

/* the main program that runs in the background, playing all the music */


#include "sound.h"

int audio, sock, iph=-1;
song_chain *songlist=NULL;

#define GRDL_VER "v1.0"


void main(argc, argv)
int argc;
char *argv[];
{
  int a, l, itr=0, opt=0, chap=0, snum=0;
  char *cmd=*argv;

  /* grab any flags on the command line */
  while(argc > 1) {
    --argc; ++argv;
    if(opt) {
      if(opt & 1)
        chap=atoi(*argv);
      if(opt & 2)
        snum=atoi(*argv);
      opt=0;
      continue;
    }
    if(**argv == '-') {
      for(a=1;(*argv)[a];a++) {
        switch((*argv)[a]) {
        case 'c':
          opt|=1;
          break;
        case 'i':
          itr=1;
          break;
        case 's':
          opt|=2;
          break;
        case 'v':
          fprintf(stdout, "%s: GRDL Version Reference %s\n", cmd, GRDL_VER);
          fflush(stdout);
          exit(0);
        default:
          fprintf(stderr, "%s: illegal option -- %c\n", cmd, (*argv)[a]);
          fprintf(stderr, "Usage: %s [-iv] [-c chapter] [-s song]\n", cmd);
          fflush(stderr);
          exit(1);
        }
      }
      continue;
    }
  }

  if(!itr) {
    /* close standard input */
    fclose(stdin);
    if(fork() != 0)
      exit(0);
    setpgrp();
  }
  signal(SIGTERM, SIG_IGN);

  /* open the audio file */
  if((audio = open("/dev/dsp", O_WRONLY, 0)) == -1) {
    perror("/dev/dsp");
    exit(1);
  }

  /* open the input socket */
  sock=open_socket();

  /* load in the startup tune */
  read_song((chap*256)+snum);

#define SONGEND ((songlist->repeat_mode == 3 && songlist->repeat > 0L) ? \
                 songlist->repeat:songlist->length)

  while(1) {
    /* wait for song, if none present */
    if(!songlist || songlist->pause) {
      sleep(1);
      read_input();
      continue;
    }

    /* if we're repeating the song, start somewhere in the middle */
    if(songlist->c >= SONGEND) {
      songlist->c = (songlist->repeat_mode>=2)?0L:songlist->repeat;
      if(songlist->scale != -1) {
        l=speed_ctrl(0);
        if(songlist->scale == l)
          songlist->scale = -1;
        else {
          if(songlist->scale < l)
            l--;
          else
            l++;
          ioctl(audio, SNDCTL_DSP_SYNC);
          speed_ctrl(l);
          continue;
        }
      }
      if(songlist->fpause) {
        ioctl(audio, SNDCTL_DSP_SYNC);
        read_input();
        continue;
      }
      if(songlist->repeat_num > -2)
        --songlist->repeat_num;
      if(!songlist->repeat_mode || songlist->repeat_num == -1) {
        song_chain *new;

        /* free up song */
        new=songlist->next;
        if(songlist->fd != -1)
          close(songlist->fd);
        free(songlist->data);
        free(songlist);

        songlist=new;
        ioctl(audio, SNDCTL_DSP_SYNC);
        if(new)
          set_audio(new);
        continue;
      }
    }

    /* play song until end is reached */
    while(1) {
      /* check for sound controls */
      if(!songlist->fpause)
        read_input();
      if(!songlist || songlist->pause)
        break;

      /* load some more of the song */
      /* note that song reads from file while it is playing. this prevents
       * long halt before song starts while reading everything from disk */

      if(songlist->fd != -1) {
        songlist->ldc += read(songlist->fd, songlist->data+songlist->ldc,
             (songlist->length-songlist->ldc)>
             (!songlist->ldc?4096:32768) ?
             (!songlist->ldc?4096:32768) : songlist->length-songlist->ldc);
        if(songlist->ldc>=songlist->length) {
          close(songlist->fd);
          songlist->fd=-1;
        }
      }

      /* write out 4096 more bytes of the song */
      l = SONGEND-songlist->c;
      if(l > 4096)
        l = 4096;

      songlist->c += write(audio, songlist->data+songlist->c, l);

      if(songlist->c >= SONGEND)
        break;
    }
  }
}

int read_song(track)
int track;
{
  song_chain *new;
  wave_header head;
  char soundfile[256];
  int fd;

  /* read new song */
  sprintf(soundfile, "%s/ch%02x/%02x.wav", SOUND_LIBRARY, track/256, track%256);
  if((fd = open(soundfile, O_RDONLY, 0)) == -1)
    return 0;

  /* open new song pointer */
  new=(song_chain *)malloc(sizeof(song_chain));
  new->next = songlist;
  songlist=new;

  /* read header from file */
  read(fd, &head, sizeof(head));
  new->stereo = (head.modus == 2);
  new->speed = head.sample_fq;
  new->sample_size = head.bit_p_spl;
  new->length = head.data_length-4;
  new->repeat = head.data_repeat;
  if(new->repeat > new->length)
    new->repeat=0L;

  new->default_speed = new->speed;
  new->repeat_mode = 1;
  new->repeat_num = -1;
  new->pause = 0;
  new->fpause = 0;
  new->scale = -1;

  new->tune = track;
  new->data = (void *)malloc(new->length);
  new->fd = fd;
  new->c = 0;
  new->ldc = 0;

  set_audio(new);
  return 1;
}

void set_audio(song)
song_chain *song;
{
  int a;

  /* clear existing sound audio */
  ioctl(audio, SNDCTL_DSP_RESET);

  ioctl(audio, SNDCTL_DSP_SAMPLESIZE, &song->sample_size);
  ioctl(audio, SNDCTL_DSP_STEREO, &song->stereo);

  /* set speed and sound driver accessibility */
  a=song->speed;
  if(ioctl(audio, SNDCTL_DSP_SPEED, &a) == -1) {
    perror("/dev/dsp");
    exit(1);
  }
}

int open_socket() 
{
  int s;

  /* open the socket */
  if ((s = open("/proc/snddriv", O_RDONLY|O_NDELAY)) < 0) {
    perror("open");
    exit(1);
  }

  return s;
}

void read_input()
{
  song_chain *new, *snext;
  static int pc=0, queue=0;
  static unsigned char port[3];
  int l, arg=0;

  if(songlist && songlist->fpause)
    songlist->fpause=0;

  if(queue >= 0)
    queue--;
  /* process input data */
  while((l=read(sock, port+pc, 1)) == 1) {
    queue=2;
    if(!songlist && (*port != WV_SELECT && *port != WV_NEXT &&
                     *port != WV_ALLOC)) {
      pc=0;
      continue;
    } else if(songlist && songlist->pause)
      songlist->pause=2;

    switch(*port) {
    case WV_PING:
      set_audio(songlist);
      pc=-1;
      break;
    case WV_DEFAULT:
      songlist->repeat_mode=1;
      pc=-1;
      break;
    case WV_SPEED:
      if(pc == 1) {
        songlist->speed=*(port+1)*1000;
        set_audio(songlist);
        pc=-1;
      }
      break;
    case WV_NORMAL:
      songlist->speed=songlist->default_speed;
      set_audio(songlist);
      pc=-1;
      break;
    case WV_FREE:
      if(!songlist->next) {
        pc=-1;
        break;
      }
      new=songlist->next;
      if(songlist->fd != -1)
        close(songlist->fd);
      free(songlist->data);
      free(songlist);

      songlist=new;
      if(new)
        set_audio(new);
      pc=-1;
      break;
    case WV_ALLOC:
      if(pc == 2 && songlist) {
        arg=*(port+2);
        arg+=(*(port+1))*256;
        read_song(arg);
        pc=-1;
        break;
      } else if(pc < 2)
        break;
    case WV_SELECT:
    case WV_PREV:
    case WV_NEXT:
    case WV_OFF:
    case WV_RESTART:
    case WV_MIDWAY:
      if(*port != WV_SELECT)
        pc=2;
      if(pc == 2) {
        if(*port != WV_SELECT && *port != WV_ALLOC) {
          arg=0;
        } else {
          arg=*(port+2);
          arg+=(*(port+1))*256;
        }
        for(new=songlist;new;new=snext) {
          if(!(snext=new->next))
            break;
          if(new->fd != -1)
            close(new->fd);
          free(new->data);
          free(new);
        }
        if(*port != WV_SELECT && *port != WV_ALLOC) {
          if(!new)
            arg = 0;
          else
            arg = new->tune;
          if(*port == WV_PREV)
            arg--;
          if(*port == WV_NEXT)
            arg++;
          if(*port == WV_OFF || arg < 0 || arg > 65535)
            arg=0;
        }
        if(new) {
          if(arg == new->tune) {
            if(new != songlist) {
              set_audio(new);
              songlist=new;
            } else if(*port == WV_RESTART || *port == WV_MIDWAY)
              ioctl(audio, SNDCTL_DSP_RESET);
            if(*port == WV_RESTART)
              songlist->c=0L;
            else if(*port == WV_MIDWAY)
              songlist->c=songlist->repeat;
            pc=-1;
            break;
          }
          if(new->fd != 1)
            close(new->fd);
          free(new->data);
          free(new);
          songlist=NULL;
        }
        read_song(arg);
        pc=-1;
      }
      break;
    case WV_REPEATOFF:
      songlist->repeat_mode=0;
      pc=-1;
      break;
    case WV_REPEATON:
      songlist->repeat_mode=1;
      pc=-1;
      break;
    case WV_REPEATFULL:
      songlist->repeat_mode=2;
      pc=-1;
      break;
    case WV_REPEATMIDDLE:
      songlist->repeat_mode=3;
      pc=-1;
      break;
    case WV_REPEATNUM:
      if(pc == 1) {
        songlist->repeat_num=*(port+1);
        pc=-1;
      }
      break;
    case WV_REPEATINF:
      songlist->repeat_num=-1;
      pc=-1;
      break;
    case WV_SCALE:
      if(pc == 1) {
        songlist->scale=*(port+1);
        if(songlist->scale < 17)
          songlist->scale = 17;
        if(songlist->scale > 30)
          songlist->scale = 30;
        pc=-1;
      }
      break;
    case WV_PAUSE:
      if(!songlist->pause)
        songlist->pause=1;
      pc=-1;
      break;
    case WV_WAIT:
      if(pc == 1) {
        sleep(*(port+1));
        pc=-1;
      }
      break;
    case WV_RESET:
      ioctl(audio, SNDCTL_DSP_RESET);
      pc=-1;
      break;
    case WV_SYNC:
      ioctl(audio, SNDCTL_DSP_SYNC);
      pc=-1;
      break;
    case WV_FINISH:
      songlist->fpause=1;
      pc=-1;
      break;
    default:
      if(*port > 16 && *port < 31)
        speed_ctrl(*port);
      pc=-1;
    }
    pc++;
    if(songlist && songlist->fpause)
      break;
  }

  if(!queue) {
    printf("[v");
    fflush(stdout);
    lseek(sock, 0L, SEEK_SET);
  }

  if(songlist && songlist->pause == 2)
    songlist->pause=0;
}

int speed_ctrl(lvl)
int lvl;
{
  int a, speeds[]={2205,4410,8820,11025,13230,15435,17640,22050,26460,
                   30870,35280,39690,44100,48510};

  if(!lvl) {
    for(a=0;a<13;a++) {
      if(songlist->speed < speeds[a+1])
        return a+17;
    }
    return 30;
  }

  if(lvl < 17)
    lvl=17;
  else if(lvl > 30)
    lvl=30;
  if(songlist->speed != speeds[lvl-17]) {
    songlist->speed=speeds[lvl-17];
    set_audio(songlist);
  }

  return lvl;
}
