/*  Copyright (C) 1995  Byron Stanoszek
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 **************************************************************************/

/* sound.h */


#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <fcntl.h>
#include <signal.h>
#include <unistd.h>
#include <sys/soundcard.h>
#include <sys/wav.h>


/* header for WAV-Files */
typedef struct {
  char main_chunk[4];	/* 'RIFF' */
  long length;		/* length of file */
  char chunk_type[4];	/* 'WAVE' */
  char sub_chunk[4];	/* 'fmt ' */
  long length_chunk;	/* length sub_chunk, always 16 bytes */
  short format;		/* always 1 = PCM-Code */
  short modus;		/* 1 = Mono, 2 = Stereo */
  long sample_fq;	/* Sample Freq */
  long byte_p_sec;	/* Data per sec */
  short byte_p_spl;	/* bytes per sample:
					1=8 bit, 2=16 bit (mono)
					2=8 bit, 4=16 bit (stereo) */
  short bit_p_spl;	/* bits per sample, 8, 12, 16 */
  char data_chunk[4];	/* 'data' */
  long data_length;	/* length of data */
  long data_repeat;	/* repeat point */
} wave_header;

typedef struct song_chain_struct song_chain;
struct song_chain_struct {
  song_chain *next;

  int tune;		/* tune number */
  char *data;		/* song data */
  int c;		/* play counter */
  int ldc;		/* load counter */
  int fd;		/* file descriptor */

  int stereo;		/* 1 = Mono, 2 = Stereo */
  long speed;		/* Data per sec */
  short sample_size;	/* bits per sample */
  long length;		/* length of file */
  long repeat;		/* location of file which repeats */

  int default_speed;	/* default speed of tune */
  int repeat_mode;	/* 0 = Single, 1 = Repeat, 2 = Full, 3 = Middle */
  int repeat_num;	/* how many times still to repeat */
  int fpause;		/* halt data codes until song finishes playing */
  int pause;		/* pause mode */
  int scale;		/* scale speed */
};

/* global declarations */
int atoi();
int read();
int ioctl();
void *malloc();
void free();
int write();
int read_song();
void set_audio();
int open_socket();
int unlink();
void read_input();
void send_data();
int fork();
int setpgrp();
int speed_ctrl();
